module Roots

using StaticArrays
using LinearAlgebra

export Root, isproportional, isorthogonal, ~, ⟂

abstract type AbstractRoot end

struct Root{N} <: AbstractRoot
    coord::SVector{N, Float64}
    
    Root(a::SVector{N}) where N = new{N}(a)
end

Root(a) = Root(SVector(a...))

function Base.:(==)(r::Root{N}, s::Root{M}) where {M, N}
    M == N || return false
    r.coord == s.coord || return false
    return true
end

Base.hash(r::Root, h::UInt) = hash(Root, hash(r.coord, h))

Base.:+(r::Root{N}, s::Root{N}) where N = Root(r.coord+s.coord)
Base.:-(r::Root{N}, s::Root{N}) where N = Root(r.coord-s.coord)
Base.:-(r::Root{N}) where N = Root(-r.coord)

Base.:*(a::Number, r::Root) = Root(a*r.coord)
Base.:*(r::Root, a::Number) = a*r

Base.length(r::Root) = norm(r, 2)

LinearAlgebra.norm(r::Root, p::Real=2) = norm(r.coord, p)

LinearAlgebra.dot(r::Root{N}, s::Root{N}) where N = dot(r.coord, s.coord)

cos_angle(a,b) = dot(a, b) / (norm(a) * norm(b))

function isproportional(α::Root{N}, β::Root{N}) where N 
    return isapprox(abs(cos_angle(α, β)), 1.0, atol=eps(1.0))
end

Base.:~(α::Root{N}, β::Root{N}) where N = isproportional(α, β)

function isorthogonal(α::Root{N}, β::Root{N}) where N
    return isapprox(cos_angle(α, β), 0.0, atol=eps(1.0))
end

⟂(α::Root{N}, β::Root{N}) where N = isorthogonal(α, β)

function Base.show(io::IO, r::Root{N}) where N
    print(io, "$(r.coord): root of length $(norm(r))")
end

E(N, i::Integer) = Root(ntuple(k -> k == i ? 1 : 0, N))

end # of module RootSystems
