module SpNs

using LinearAlgebra
using AbstractAlgebra
using Nemo
using Groups
using GroupRings

include("Roots.jl")
using .Roots

export Sp, isproportional, ~, isorthogonal, ⟂, positive

# two families of generators

function a(i::Int, j::Int, M::MatSpace, val=one(M.base_ring)) # short roots
    @assert i ≠ j
    m = one(M)
    N = size(m,2)÷2
    
    m[i,j] = val
    m[N+j, N+i] = -val
    @assert issymplectic(m)
    return m
end

function b(i::Int, j::Int, M::MatSpace, val=one(M.base_ring)) # long roots
    m = one(M)
    N = size(m,2)÷2
    m[i, N+j] = val
    m[j, N+i] = val
    @assert issymplectic(m)
    return m
end

# symplectic generator with associated root datum:

struct Sp{N}
    root::Root{N}
    matrix::Nemo.fmpz_mat
end

function Sp{N}(typ::Val{:a}, i::Int, j::Int) where N
    @assert i≠j
    r = Roots.E(N, i) - Roots.E(N, j)
    M = Nemo.MatrixSpace(Nemo.ZZ, 2N, 2N)
    m = a(i,j, M)
    return Sp{N}(r, m)
end

function Sp{N}(typ::Val{:b}, i::Int, j::Int=i) where N
    r = Roots.E(N, i) + Roots.E(N, j)
    M = Nemo.MatrixSpace(Nemo.ZZ, 2N, 2N)
    m = b(i,j, M)
    return Sp{N}(r, m)
end

Sp{N}(s::Symbol, i::Int, j::Int=i) where N = Sp{N}(Val(s), i, j)

Base.inv(s::Sp) = Sp(s.root, inv(s.matrix))
Base.:-(s::Sp) = Sp(-s.root, transpose(s.matrix))

Roots.isproportional(x::Sp, y::Sp) = isproportional(x.root, y.root)
Roots.:~(x::Sp,y::Sp) = x.root ~ y.root

Roots.isorthogonal(x::Sp, y::Sp) = isorthogonal(x.root, y.root)
Roots.:⟂(x::Sp,y::Sp) = x.root ⟂ y.root

function issymplectic(M)
    r = nrows(M)
    c = ncols(M)
    @assert r == c
    @assert iseven(r)
    n = r÷2
    I = Diagonal(ones(Int, n))
    
    O = zeros(Int,n,n);
    Ω = parent(M)([O  I; -I O])
    return Ω == transpose(M)*Ω*M
end

indexing(n) = [(i,j) for i in 1:n for j in i+1:n]

function gens_roots(N)
    a_ijs = [Sp{N}(:a, i,j) for (i,j) in indexing(N)]
    b_is = [Sp{N}(:b, i) for i in 1:N]
    c_ijs = [Sp{N}(:b, i,j) for (i,j) in indexing(N)]
    root_system = [a_ijs; b_is; c_ijs]
    root_system = [root_system; [-r for r in root_system]];
    root_system = [root_system; inv.(root_system)]
    return root_system
end

function positive(root_system::Vector{SP}) where SP
    roots = [r.root for r in root_system]
    
    pos_roots = eltype(roots)[]
    for (idx,γ) in enumerate(roots)
        if any(isproportional(s, γ) for s in roots[1:idx-1])
            continue
        end
        push!(pos_roots, γ)
        @info "" positive_root = γ
    end
    # TODO: what about their sums??
    
    return pos_roots
end

function laplacians(RG::GroupRing, root_system::Vector{SpNs.Sp{N}}) where N
    positive_roots = positive(root_system)
    Sαs = Dict(α => [w.matrix for w in root_system if isproportional(w.root, α)] for α in positive_roots)
    Δs = Dict(α => RG(length(S)) - RG(S) for (α, S) in Sαs)
    return Δs
end

include("SpN_actions.jl")

end # of module SPNs
